/*
  UOC - Treball Final de Grau
  @file OpenDCert.sol
  @desc Smart-contract per l'aplicació descentralitzada OpenDCert.

  @author     Daniel Atienza Lopez
  @version    1.0.0
  @date       23/04/2018
*/

pragma solidity ^0.4.20;

contract OpenDCert {

  struct Organization {
    string ipfs_hash;
    uint16 qt_certificates_issued;
    address added_by;
    uint256 added_timestamp;
    uint256 updated_timestamp;
  }

  struct Student {
    string ipfs_hash;
    mapping (uint8 => Certificate) certificates;
    uint8 qt_certificates;
    address added_by;
    uint256 added_timestamp;
    uint256 updated_timestamp;
  }

  struct Certificate {
    uint8 id;
    string ipfs_hash;
    address added_by;
    uint256 added_timestamp;
  }

  // Constants
  bytes1 public constant T_ADMIN = 0x01;
  bytes1 public constant T_ORGANIZATION = 0x02;
  bytes1 public constant T_STUDENT = 0x03;
  bytes10 public constant VERSION = "1.0.0";

  address public _owner;
  uint256 public _created_timestamp;
  uint256 public _qt_users = 0;

  // Users mapped to the user type 1 - Admin, 2 - Organizations (Schools, Universities, ...), 3 - Student
  mapping (address => bytes1) public _users;
  mapping (address => Organization) public _organizations;
  mapping (address => Student) public _students;

  /*
    constructor
  */
  constructor() public payable
  {
    _owner = msg.sender;
    _created_timestamp = block.timestamp;

    _users[msg.sender] = T_ADMIN;
    _qt_users++;
  }

  /*
    addUser:
    Method to insert a new organization or a new student.
    The administrator only can add organizations, and the organizations only
    can add students.

    @param address  paddress  Clau publica del usuari
  */
  function addUser(address paddress) public payable
  {
    require((_users[msg.sender] == T_ADMIN || _users[msg.sender] == T_ORGANIZATION) && _users[paddress] == 0x00);

    bytes1 user_type;

    if (_users[msg.sender] == T_ADMIN) {

      user_type = T_ORGANIZATION;

      Organization memory objOrganization;

      objOrganization.ipfs_hash = "";
      objOrganization.qt_certificates_issued = 0;
      objOrganization.added_by = msg.sender;
      objOrganization.added_timestamp = block.timestamp;

      _organizations[paddress] = objOrganization;
    }
    else if (_users[msg.sender] == T_ORGANIZATION) {

      user_type = T_STUDENT;

      Student memory objStudent;

      objStudent.ipfs_hash = "";
      objStudent.qt_certificates = 0;
      objStudent.added_by = msg.sender;
      objStudent.added_timestamp = block.timestamp;

      _students[paddress] = objStudent;
    }

    _users[paddress] = user_type;
    _qt_users += 1;
  }

  /*
    addCertificate:
    Method to insert a new certificate of a student.
    Only organizations can add certificates.

    @param address  pstudent    Clau publica de l'alumne
    @param string   pipfs_hash  Hash del fitxer en el sistema IPFS
  */
  function addCertificate(address pstudent, string pipfs_hash) public payable
  {
    require(_users[msg.sender] == T_ORGANIZATION && _users[pstudent] == T_STUDENT);

    Student storage objStudent = _students[pstudent];
    Certificate memory objCertificate;

    objCertificate.id = objStudent.qt_certificates + 1;
    objCertificate.ipfs_hash = pipfs_hash;
    objCertificate.added_by = msg.sender;
    objCertificate.added_timestamp = block.timestamp;

    objStudent.certificates[objStudent.qt_certificates] = objCertificate;
    objStudent.qt_certificates += 1;

    Organization storage objOrganization = _organizations[msg.sender];
    objOrganization.qt_certificates_issued += 1;

    _students[pstudent] = objStudent;
    _organizations[msg.sender] = objOrganization;
  }

  /*
    setStudent / setOrganization:
    Method to update the information of a organization or a student.
    Only the organization or the student can update their details.

    @param string pipfs_hash IPFS hash with the details of the student
  */
  function setStudent(string pipfs_hash) public payable
  {
    require(_users[msg.sender] == T_STUDENT);

    Student memory objStudent = _students[msg.sender];

    objStudent.ipfs_hash = pipfs_hash;
    objStudent.updated_timestamp = block.timestamp;

    _students[msg.sender] = objStudent;
  }

  function setOrganization(string pipfs_hash) public payable
  {
    require(_users[msg.sender] == T_ORGANIZATION);

    Organization memory objOrganization = _organizations[msg.sender];

    objOrganization.ipfs_hash = pipfs_hash;
    objOrganization.updated_timestamp = block.timestamp;

    _organizations[msg.sender] = objOrganization;
  }

  /*
    getCertificate:
    Method to get a specific certificate of a user.

    @param address pstudent Clau publica de l'alumne

    @return address Organization address.
    @return string  IPFS hash with the details of the certificate.
    @return uint256 Timestamp when the certificate was added.
  */
  function getCertificate(address pstudent, uint8 pcertificate) public constant returns (uint8, address, string, uint256)
  {
    require(_users[pstudent] == T_STUDENT);
    Certificate memory objCertificate = _students[pstudent].certificates[pcertificate];
    return (objCertificate.id, objCertificate.added_by, objCertificate.ipfs_hash, objCertificate.added_timestamp);
  }
}
